"""Streaming callbacks init."""

from .history import History
from .msprt import mSPRT
from .warning_samples import WarningSamplesBuffer

__all__ = [
    "History",
    "mSPRT",
    "WarningSamplesBuffer",
]



"""Decorators module."""

from functools import wraps
from inspect import signature
from typing import Any

from frouros.common.exceptions import ArgumentError


def check_func_parameters(func: Any) -> Any:
    """Decorator function to check function parameters.

    :param func: function to be wrapped
    :type func: Any
    :return: wrapper
    :rtype: Any
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        func_parameters = signature(args[1]).parameters
        if "y_true" not in func_parameters or "y_pred" not in func_parameters:
            raise ArgumentError("value function must have y_true and y_pred arguments.")
        func(*args, **kwargs)

    return wrapper



"""ECDD (EWMA for Concept Drift Detection) module."""

from typing import List, Optional, Union

import numpy as np  # type: ignore

from frouros.callbacks.base import BaseCallback
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPC,
    BaseECDDConfig,
)
from frouros.utils.stats import EWMA, Mean


class ECDDWTConfig(BaseECDDConfig):
    """ECDDWT (EWMA Concept Drift Detection Warning) [ross2012exponentially]_ configuration.

    :References:

    .. [ross2012exponentially] Ross, Gordon J., et al.
        "Exponentially weighted moving average charts for detecting concept drift."
        Pattern recognition letters 33.2 (2012): 191-198.
    """


class ECDDWT(BaseSPC):
    """ECDDWT (EWMA Concept Drift Detection Warning) [ross2012exponentially]_ detector.

    :References:

    .. [ross2012exponentially] Ross, Gordon J., et al.
        "Exponentially weighted moving average charts for detecting concept drift."
        Pattern recognition letters 33.2 (2012): 191-198.
    """

    config_type = ECDDWTConfig  # type: ignore

    def __init__(
        self,
        config: Optional[ECDDWTConfig] = None,
        callbacks: Optional[Union[BaseCallback, List[BaseCallback]]] = None,
    ) -> None:
        """Init method.

        :param config: configuration parameters
        :type config: Optional[ECDDWTConfig]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[Callback, List[BaseCallback]]]
        """
        super().__init__(
            config=config,  # type: ignore
            callbacks=callbacks,
        )
        self.additional_vars = {
            "p": Mean(),
            "z": EWMA(alpha=self.config.lambda_),  # type: ignore
            **self.additional_vars,  # type: ignore
        }
        self._set_additional_vars_callback()
        self._lambda_div_two_minus_lambda = self.config.lambda_ / (  # type: ignore
            2 - self.config.lambda_  # type: ignore
        )

    @property
    def p(self) -> Mean:
        """P property.

        :return: p value
        :rtype: Mean
        """
        return self._additional_vars["p"]

    @p.setter
    def p(self, value: Mean) -> None:
        """P setter.

        :param value: value to be set
        :type value: Mean
        :raises TypeError: Type error exception
        """
        if not isinstance(value, Mean):
            raise TypeError("value must be of type Mean.")
        self._additional_vars["p"] = value

    @property
    def z(self) -> EWMA:
        """Z property.

        :return: z value
        :rtype: Mean
        """
        return self._additional_vars["z"]

    @z.setter
    def z(self, value: EWMA) -> None:
        """Z setter.

        :param value: value to be set
        :type value: EWMA
        :raises TypeError: Type error exception
        """
        if not isinstance(value, EWMA):
            raise TypeError("value must be of type EWMA.")
        self._additional_vars["z"] = value

    def _check_threshold(
        self, control_limit: float, z_variance: float, warning_level: float = 1.0
    ) -> bool:
        return self.z.mean > self.p.mean + warning_level * control_limit * z_variance

    def reset(self) -> None:
        """Reset method."""
        super().reset()
        self.p = Mean()
        self.z = EWMA(alpha=self.config.lambda_)  # type: ignore

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1

        self.p.update(value=value)
        self.z.update(value=value)

        if self.num_instances >= self.config.min_num_instances:
            error_rate_variance = self.p.mean * (1 - self.p.mean)
            z_variance = np.sqrt(
                self._lambda_div_two_minus_lambda
                * (1 - self.z.one_minus_alpha ** (2 * self.num_instances))
                * error_rate_variance
            )
            control_limit = self.config.control_limit_func(  # type: ignore
                p=self.p.mean
            )

            drift_flag = self._check_threshold(
                control_limit=control_limit, z_variance=z_variance
            )

            if drift_flag:
                # Out-of-Control
                self.drift = True
                self.warning = False
            else:
                warning_flag = self._check_threshold(
                    control_limit=control_limit,
                    z_variance=z_variance,
                    warning_level=self.config.warning_level,  # type: ignore
                )
                if warning_flag:
                    # Warning
                    self.warning = True
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False



"""DDM (Drift detection method) module."""

from contextlib import suppress
from typing import Union

from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPCConfig,
    BaseSPCError,
)


class DDMConfig(BaseSPCConfig):
    """DDM (Drift detection method) [gama2004learning]_ configuration.

    :References:

    .. [gama2004learning] Gama, Joao, et al.
        "Learning with drift detection."
        Advances in Artificial Intelligence–SBIA 2004: 17th Brazilian Symposium on
        Artificial Intelligence, Sao Luis, Maranhao, Brazil, September 29-Ocotber 1,
        2004. Proceedings 17. Springer Berlin Heidelberg, 2004.
    """


class DDM(BaseSPCError):
    """DDM (Drift detection method) [gama2004learning]_ detector.

    :References:

    .. [gama2004learning] Gama, Joao, et al.
        "Learning with drift detection."
        Advances in Artificial Intelligence–SBIA 2004: 17th Brazilian Symposium on
        Artificial Intelligence, Sao Luis, Maranhao, Brazil, September 29-Ocotber 1,
        2004. Proceedings 17. Springer Berlin Heidelberg, 2004.
    """

    config_type = DDMConfig

    def _update(self, value: Union[int, float], **kwargs) -> None:
        self.num_instances += 1
        self.error_rate.update(value=value)

        if self.num_instances >= self.config.min_num_instances:
            error_rate_plus_std, std = self._calculate_error_rate_plus_std()

            self._update_min_values(error_rate_plus_std=error_rate_plus_std, std=std)

            drift_flag = self._check_threshold(
                error_rate_plus_std=error_rate_plus_std,
                min_error_rate=self.min_error_rate,
                min_std=self.min_std,
                level=self.config.drift_level,  # type: ignore
            )
            if drift_flag:
                # Out-of-Control
                self.drift = True
                self.warning = False
            else:
                warning_flag = self._check_threshold(
                    error_rate_plus_std=error_rate_plus_std,
                    min_error_rate=self.min_error_rate,
                    min_std=self.min_std,
                    level=self.config.warning_level,  # type: ignore
                )
                if warning_flag:
                    # Warning
                    self.warning = True
                    for callback in self.callbacks:  # type: ignore
                        with suppress(AttributeError):
                            callback.on_warning_detected(**kwargs)  # type: ignore
                else:
                    # In-Control
                    self.warning = False
                self.drift = False
        else:
            self.drift, self.warning = False, False



"""Concept drift CUSUM based detection methods' init."""

from .cusum import CUSUM, CUSUMConfig
from .geometric_moving_average import (
    GeometricMovingAverage,
    GeometricMovingAverageConfig,
)
from .page_hinkley import PageHinkley, PageHinkleyConfig

__all__ = [
    "CUSUM",
    "CUSUMConfig",
    "GeometricMovingAverage",
    "GeometricMovingAverageConfig",
    "PageHinkley",
    "PageHinkleyConfig",
]



"""Batch callbacks init."""

from .permutation_test import PermutationTestOnBatchData
from .reset_drift import ResetOnBatchDataDrift

__all__ = [
    "PermutationTestOnBatchData",
    "ResetOnBatchDataDrift",
]



"""History callback module."""

from typing import Any, Dict, List, Optional, Union

from frouros.callbacks.streaming.base import BaseCallbackStreaming
from frouros.utils.stats import BaseStat


class History(BaseCallbackStreaming):
    """History callback class."""

    def __init__(self, name: Optional[str] = None) -> None:
        """Init method.

        :param name: name value
        :type name: Optional[str]
        """
        super().__init__(name=name)
        self.additional_vars: List[str] = []
        self.history: Dict[str, List[Any]] = {
            "value": [],
            "num_instances": [],
            "drift": [],
        }

    def add_additional_vars(self, vars_: List[str]) -> None:
        """Add addtional variables to track.

        :param vars_: list of variables
        :type vars_: List[str]
        """
        self.additional_vars.extend(vars_)
        self.history = {**self.history, **{var: [] for var in self.additional_vars}}

    def on_update_end(self, value: Union[int, float], **kwargs) -> None:
        """On update end method.

        :param value: value to update detector
        :type value: int
        """
        self.history["value"].append(value)
        self.history["num_instances"].append(
            self.detector.num_instances  # type: ignore
        )
        self.history["drift"].append(self.detector.drift)  # type: ignore
        for var in self.additional_vars:
            additional_var = self.detector.additional_vars[var]  # type: ignore
            # FIXME: Extract isinstance check to be done when  # pylint: disable=fixme
            #  add_addtional_vars is called (avoid the same computation)
            self.history[var].append(
                additional_var.get()
                if isinstance(additional_var, BaseStat)
                else additional_var
            )

        self.logs.update(**self.history)

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # def set_detector(self, detector) -> None:
    #     """Set detector method.
    #
    #     :raises TypeError: Type error exception
    #     """
    #     if not isinstance(detector, BaseConceptDrift):
    #         raise TypeError(
    #             f"callback {self.__class__.name} cannot be used with detector"
    #             f" {detector.__class__name}. Must be used with a detector of "
    #             f"type BaseConceptDrift."
    #         )
    #     self.detector = detector

    def reset(self) -> None:
        """Reset method."""
        for key in self.history.keys():
            self.history[key].clear()



"""Common exception module."""


class ArgumentError(ValueError):
    """Argument exception."""


class OneSampleError(Exception):
    """One sample exception."""


class UpdateDetectorError(Exception):
    """Update detector exception."""



"""MMD (Maximum Mean Discrepancy) module."""

import itertools
import math
from typing import Callable, Generator, Optional, List, Union

import numpy as np  # type: ignore
import tqdm  # type: ignore
from scipy.spatial.distance import cdist  # type: ignore

from frouros.callbacks.base import BaseCallback
from frouros.detectors.data_drift.base import MultivariateData
from frouros.detectors.data_drift.batch.distance_based.base import (
    BaseDistanceBased,
    DistanceResult,
)


def rbf_kernel(
    X: np.ndarray, Y: np.ndarray, std: float = 1.0  # noqa: N803
) -> np.ndarray:
    """Radial basis function kernel between X and Y matrices.

    :param X: X matrix
    :type X: numpy.ndarray
    :param Y: Y matrix
    :type Y: numpy.ndarray
    :param std: standard deviation value
    :type std: float
    :return: Radial basis kernel matrix
    :rtype: numpy.ndarray
    """
    return np.exp(-cdist(X, Y, "sqeuclidean") / 2 * std**2)


class MMD(BaseDistanceBased):
    """MMD (Maximum Mean Discrepancy) [gretton2012kernel]_ detector.

    :References:

    .. [gretton2012kernel] Gretton, Arthur, et al.
        "A kernel two-sample test."
        The Journal of Machine Learning Research 13.1 (2012): 723-773.
    """

    def __init__(
        self,
        kernel: Callable = rbf_kernel,
        chunk_size: Optional[int] = None,
        callbacks: Optional[Union[BaseCallback, List[BaseCallback]]] = None,
    ) -> None:
        """Init method.

        :param kernel: kernel function
        :type kernel: Callable
        :param chunk_size: chunk size value
        :type chunk_size: Optional[int]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[BaseCallback, List[BaseCallback]]]
        """
        super().__init__(
            statistical_type=MultivariateData(),
            statistical_method=self._mmd,
            statistical_kwargs={
                "kernel": kernel,
            },
            callbacks=callbacks,
        )
        self.kernel = kernel
        self.chunk_size = chunk_size
        self._chunk_size_x = None
        self.X_chunks_combinations = None
        self.X_num_samples = None
        self.expected_k_xx = None

    @property
    def chunk_size(self) -> Optional[int]:
        """Chunk size property.

        :return: chunk size to use
        :rtype: int
        """
        return self._chunk_size

    @chunk_size.setter
    def chunk_size(self, value: Optional[int]) -> None:
        """Chunk size method setter.

        :param value: value to be set
        :type value: Optional[int]
        :raises TypeError: Type error exception
        """
        if value is not None:
            if isinstance(value, int):  # type: ignore
                if value <= 0:
                    raise ValueError("chunk_size must be greater than 0 or None.")
            else:
                raise TypeError("chunk_size must be of type int or None.")
        self._chunk_size = value

    @property
    def kernel(self) -> Callable:
        """Kernel property.

        :return: kernel function to use
        :rtype: Callable
        """
        return self._kernel

    @kernel.setter
    def kernel(self, value: Callable) -> None:
        """Kernel method setter.

        :param value: value to be set
        :type value: Callable
        :raises TypeError: Type error exception
        """
        if not isinstance(value, Callable):  # type: ignore
            raise TypeError("kernel must be of type Callable.")
        self._kernel = value

    def _distance_measure(
        self,
        X_ref: np.ndarray,  # noqa: N803
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> DistanceResult:
        mmd = self._mmd(X=X_ref, Y=X, kernel=self.kernel, **kwargs)
        distance_test = DistanceResult(distance=mmd)
        return distance_test

    def _fit(
        self,
        X: np.ndarray,  # noqa: N803
        **kwargs,
    ) -> None:
        super()._fit(X=X)
        # Add dimension only for the kernel calculation (if dim == 1)
        if X.ndim == 1:
            X = np.expand_dims(X, axis=1)  # noqa: N806
        self.X_num_samples = len(self.X_ref)  # type: ignore # noqa: N806

        self._chunk_size_x = (
            self.X_num_samples
            if self.chunk_size is None
            else self.chunk_size  # type: ignore
        )

        X_chunks = self._get_chunks(  # noqa: N806
            data=X,
            chunk_size=self._chunk_size_x,  # type: ignore
        )
        xx_chunks_combinations = itertools.product(X_chunks, repeat=2)  # noqa: N806

        if kwargs.get("verbose", False):
            num_chunks = (
                math.ceil(self.X_num_samples / self._chunk_size_x) ** 2  # type: ignore
            )
            xx_chunks_combinations = tqdm.tqdm(
                xx_chunks_combinations,
                total=num_chunks,
            )

        k_xx_sum = (
            self._compute_kernel(
                chunk_combinations=xx_chunks_combinations,  # type: ignore
                kernel=self.kernel,
            )
            # Remove diagonal (j!=i case)
            - self.X_num_samples  # type: ignore
        )

        self.expected_k_xx = k_xx_sum / (  # type: ignore
            self.X_num_samples * (self.X_num_samples - 1)  # type: ignore
        )

    @staticmethod
    def _compute_kernel(chunk_combinations: Generator, kernel: Callable) -> float:
        k_sum = np.array([kernel(*chunk).sum() for chunk in chunk_combinations]).sum()
        return k_sum

    @staticmethod
    def _get_chunks(data: np.ndarray, chunk_size: int) -> Generator:
        chunks = (
            data[i : i + chunk_size]  # noqa: E203
            for i in range(0, len(data), chunk_size)
        )
        return chunks

    def _mmd(  # pylint: disable=too-many-locals
        self,
        X: np.ndarray,  # noqa: N803
        Y: np.ndarray,
        *,
        kernel: Callable,
        **kwargs,
    ) -> float:  # noqa: N803
        # Only check for X dimension (X == Y dim comparison has been already made)
        if X.ndim == 1:
            X = np.expand_dims(X, axis=1)  # noqa: N806
            Y = np.expand_dims(Y, axis=1)  # noqa: N806

        X_chunks = self._get_chunks(  # noqa: N806
            data=X,
            chunk_size=self._chunk_size_x,  # type: ignore
        )
        y_num_samples = len(Y)  # noqa: N806
        chunk_size_y = y_num_samples if self.chunk_size is None else self.chunk_size
        y_chunks, y_chunks_copy = itertools.tee(  # noqa: N806
            self._get_chunks(
                data=Y,
                chunk_size=chunk_size_y,  # type: ignore
            ),
            2,
        )
        y_chunks_combinations = itertools.product(  # noqa: N806
            y_chunks,
            repeat=2,
        )
        xy_chunks_combinations = itertools.product(  # noqa: N806
            X_chunks,
            y_chunks_copy,
        )

        if kwargs.get("verbose", False):
            num_chunks_y = math.ceil(y_num_samples / chunk_size_y)  # type: ignore
            num_chunks_y_combinations = num_chunks_y**2
            num_chunks_xy = (
                math.ceil(len(X) / self._chunk_size_x) * num_chunks_y  # type: ignore
            )
            y_chunks_combinations = tqdm.tqdm(
                y_chunks_combinations,
                total=num_chunks_y_combinations,
            )
            xy_chunks_combinations = tqdm.tqdm(
                xy_chunks_combinations,
                total=num_chunks_xy,
            )

        k_yy_sum = (
            self._compute_kernel(
                chunk_combinations=y_chunks_combinations,  # type: ignore
                kernel=kernel,
            )
            # Remove diagonal (j!=i case)
            - y_num_samples  # type: ignore
        )
        k_xy_sum = self._compute_kernel(
            chunk_combinations=xy_chunks_combinations,  # type: ignore
            kernel=kernel,
        )
        mmd = (
            self.expected_k_xx  # type: ignore
            + k_yy_sum / (y_num_samples * (y_num_samples - 1))
            - 2 * k_xy_sum / (self.X_num_samples * y_num_samples)  # type: ignore
        )
        return mmd



"""Base callback batch module."""

import abc

from frouros.callbacks.base import BaseCallback


class BaseCallbackBatch(BaseCallback):
    """Callback batch class."""

    def on_compare_start(self) -> None:
        """On compare start method."""

    def on_compare_end(self, **kwargs) -> None:
        """On compare end method."""

    # FIXME: set_detector method as a workaround to  # pylint: disable=fixme
    #  avoid circular import problem. Make it an abstract method and
    #  uncomment commented code when it is solved

    # @abc.abstractmethod
    # def set_detector(self, detector) -> None:
    #     """Set detector method."""

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset method."""



"""Data drift batch statistical test detection methods' init."""

from .chisquare import ChiSquareTest
from .cvm import CVMTest
from .ks import KSTest
from .welch_t_test import WelchTTest

__all__ = [
    "ChiSquareTest",
    "CVMTest",
    "KSTest",
    "WelchTTest",
]



"""MMD (Maximum Mean Discrepancy) module."""

from typing import Any, Callable, Dict, Optional, List, Tuple, Union

import numpy as np  # type: ignore

from frouros.callbacks.base import BaseCallback
from frouros.detectors.data_drift.base import NumericalData, MultivariateData
from frouros.detectors.data_drift.batch import MMD as MMDBatch  # noqa: N811
from frouros.detectors.data_drift.batch.distance_based.mmd import rbf_kernel
from frouros.detectors.data_drift.streaming.distance_based.base import (
    BaseDistanceBased,
    DistanceResult,
)
from frouros.utils.data_structures import CircularQueue


class MMD(BaseDistanceBased):
    """MMD (Maximum Mean Discrepancy) [gretton2012kernel]_ detector.

    :References:

    .. [gretton2012kernel] Gretton, Arthur, et al.
        "A kernel two-sample test."
        The Journal of Machine Learning Research 13.1 (2012): 723-773.
    """

    def __init__(
        self,
        window_size: int,
        kernel: Callable = rbf_kernel,
        chunk_size: Optional[int] = None,
        callbacks: Optional[Union[BaseCallback, List[BaseCallback]]] = None,
    ) -> None:
        """Init method.

        :param window_size: window size
        :type window_size: int
        :param kernel: kernel function
        :type kernel: Callable
        :param chunk_size: chunk size value
        :type chunk_size: Optional[int]
        :param callbacks: callbacks
        :type callbacks: Optional[Union[Callback, List[BaseCallback]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=MultivariateData(),
            callbacks=callbacks,
        )
        self.mmd = MMDBatch(
            kernel=kernel,
            chunk_size=chunk_size,
        )
        self.window_size = window_size
        self.X_queue = CircularQueue(max_len=self.window_size)

    @property
    def window_size(self) -> int:
        """Window size property.

        :return: window size
        :rtype: int
        """
        return self._window_size

    @window_size.setter
    def window_size(self, value: int) -> None:
        """Window size setter.

        :param value: value to be set
        :type value: int
        :raises ValueError: Value error exception
        """
        if value < 1:
            raise ValueError("window_size value must be greater than 0.")
        self._window_size = value

    def _fit(self, X: np.ndarray) -> None:  # noqa: N803
        self.mmd.fit(X=X)
        self.X_ref = self.mmd.X_ref

    def _reset(self) -> None:
        super()._reset()
        self.mmd.reset()

    def _update(self, value: Union[int, float]) -> Optional[DistanceResult]:
        self.X_queue.enqueue(value=value)

        if self.num_instances < self.window_size:
            return None

        # FIXME: Handle callback logs. Now are ignored.  # pylint: disable=fixme
        distance, _ = self.mmd.compare(X=np.array(self.X_queue))
        return distance

    def _compare(
        self,
        X: np.ndarray,  # noqa: N803
    ) -> Tuple[Optional[DistanceResult], Dict[str, Any]]:  # noqa: N803
        return self.mmd.compare(X=X)



"""Test callback module."""

from typing import List, Tuple

import numpy as np  # type: ignore
import pytest  # type: ignore
import sklearn  # type: ignore # pylint: disable=import-error

from frouros.callbacks.batch import (
    PermutationTestOnBatchData,
    ResetOnBatchDataDrift,
)
from frouros.callbacks.streaming import (
    History,
    mSPRT,
    WarningSamplesBuffer,
)
from frouros.detectors.concept_drift import (
    ADWIN,
    CUSUM,
    DDM,
    ECDDWT,
    EDDM,
    GeometricMovingAverage,
    HDDMA,
    HDDMW,
    KSWIN,
    PageHinkley,
    RDDM,
    STEPD,
)
from frouros.detectors.concept_drift.base import BaseConceptDrift
from frouros.detectors.concept_drift.streaming.statistical_process_control.base import (
    BaseSPC,
)
from frouros.detectors.data_drift.batch import (
    BhattacharyyaDistance,
    CVMTest,
    EMD,
    HellingerDistance,
    HINormalizedComplement,
    JS,
    KL,
    KSTest,
    MMD,
    PSI,
    WelchTTest,
)
from frouros.detectors.data_drift.batch.base import BaseDataDriftBatch
from frouros.detectors.data_drift.streaming import MMD as MMDStreaming  # noqa: N811


@pytest.mark.parametrize(
    "detector_class, expected_distance, expected_p_value",
    [
        (BhattacharyyaDistance, 0.55516059, 0.0),
        (EMD, 3.85346006, 9.21632493e-101),
        (HellingerDistance, 0.74509099, 3.13808126e-50),
        (HINormalizedComplement, 0.78, 1.31340683e-55),
        (JS, 0.67010107, 2.30485343e-63),
        (KL, np.inf, np.nan),
        (MMD, 0.69509004, 2.53277069e-137),
        (PSI, 461.20379435, 4.45088795e-238),
    ],
)
def test_batch_permutation_test_data_univariate_different_distribution(
    X_ref_univariate: np.ndarray,  # noqa: N803
    X_test_univariate: np.ndarray,
    detector_class: BaseDataDriftBatch,
    expected_distance: float,
    expected_p_value: float,
) -> None:
    """Test batch permutation test on data callback.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param detector_class: detector distance
    :type detector_class: BaseDataDriftBatch
    :param expected_distance: expected distance value
    :type expected_distance: float
    :param expected_p_value: expected p-value value
    :type expected_p_value: float
    """
    np.random.seed(seed=31)

    permutation_test_name = "permutation_test"
    detector = detector_class(  # type: ignore
        callbacks=[
            PermutationTestOnBatchData(
                num_permutations=100,
                random_state=31,
                num_jobs=-1,
                name=permutation_test_name,
            )
        ]
    )
    _ = detector.fit(X=X_ref_univariate)
    distance, callback_logs = detector.compare(X=X_test_univariate)

    assert np.isclose(distance, expected_distance)
    assert np.isclose(
        callback_logs[permutation_test_name]["p_value"],
        expected_p_value,
        equal_nan=True,
    )


@pytest.mark.parametrize(
    "detector_class",
    [CVMTest, KSTest, WelchTTest],
)
def test_batch_reset_on_data_drift(
    X_ref_univariate,  # noqa: N803
    X_test_univariate,
    detector_class: BaseDataDriftBatch,
    mocker,
) -> None:
    """Test batch reset on data drift callback.

    :param X_ref_univariate: reference univariate data
    :type X_ref_univariate: numpy.ndarray
    :param X_test_univariate: test univariate data
    :type X_test_univariate: numpy.ndarray
    :param detector_class: detector distance
    :type detector_class: BaseDataDriftBatch
    """
    mocker.patch("frouros.detectors.data_drift.batch.base.BaseDataDriftBatch.reset")

    detector = detector_class(  # type: ignore
        callbacks=[
            ResetOnBatchDataDrift(
                alpha=0.01,
            ),
        ],
    )
    _ = detector.fit(X=X_ref_univariate)
    _ = detector.compare(X=X_test_univariate)
    detector.reset.assert_called_once()  # type: ignore # pylint: disable=no-member


@pytest.mark.parametrize(
    "detector_class",
    [
        ADWIN,
        CUSUM,
        DDM,
        ECDDWT,
        EDDM,
        GeometricMovingAverage,
        HDDMA,
        HDDMW,
        KSWIN,
        PageHinkley,
        RDDM,
        STEPD,
    ],
)
def test_streaming_history_on_concept_drift(
    model_errors: List[int],
    detector_class: BaseConceptDrift,
):
    """Test streaming history on concept drift callback.

    :param model_errors: model errors
    :type model_errors: List[int]
    :param detector_class: concept drift detector
    :type detector_class: BaseConceptDrift
    """
    name = "history"
    detector = detector_class(callbacks=History(name=name))  # type: ignore

    for error in model_errors:
        history = detector.update(value=error)
        status = detector.status
        if status["drift"]:
            assert history[name]["drift"][-1]
            assert not any(history[name]["drift"][:-1])
            break


def _fit_model(model, X, y):  # noqa: N803
    model.fit(X=X, y=y)
    return model


@pytest.mark.parametrize(
    "detector_class",
    [
        DDM,
        ECDDWT,
        EDDM,
        HDDMA,
        HDDMW,
        RDDM,
    ],  # pylint: disable=too-many-locals
)
def test_streaming_warning_samples_buffer_on_concept_drift(
    dataset_simple: Tuple[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray]],
    model: sklearn.pipeline.Pipeline,
    detector_class: BaseSPC,
):
    """Test streaming warning samples buffer on concept drift callback.

    :param dataset_simple: dataset with concept drift
    :type dataset_simple: Tuple[Tuple[numpy.ndarray, numpy.ndarray],
    :param model: trained model
    :type model: sklearn.pipeline.Pipeline
    :param detector_class: concept drift detector
    :type detector_class: BaseSPC
    """
    _, test = dataset_simple  # noqa: N806

    detector = detector_class(
        callbacks=WarningSamplesBuffer(name="samples"),  # type: ignore
    )

    collect_example_warning_samples = False
    X_extra, y_extra = [], []  # noqa: N806

    for X, y in zip(*test):  # noqa: N806
        y_pred = model.predict(X.reshape(1, -1))
        if not collect_example_warning_samples:
            error = 1 - int(y_pred == y)
            callbacks_logs = detector.update(value=error, X=X, y=y)
        else:
            X_extra.append(X)
            y_extra.append(y)
        if detector.status["drift"]:
            y_new_ref = callbacks_logs["samples"]["y"] + y_extra
            if len([*set(y_new_ref)]) < 2:
                collect_example_warning_samples = True
            else:
                X_new_ref = callbacks_logs["samples"]["X"] + X_extra  # noqa: N806
                collect_example_warning_samples = False
                X_extra.clear()
                y_extra.clear()
                detector.reset()
                model = _fit_model(model=model, X=X_new_ref, y=y_new_ref)


@pytest.mark.parametrize(
    "detector_class,"
    " expected_drift_idx,"
    " expected_distance_mean,"
    " expected_p_value,"
    " expected_likelihood",
    [
        (MMDStreaming, 5, 0.27193007, 0.00989585, 101.05240437),
    ],
)
def test_streaming_msprt_multivariate_different_distribution(
    X_ref_multivariate: np.ndarray,  # noqa: N803
    X_test_multivariate: np.ndarray,
    detector_class: BaseDataDriftBatch,
    expected_drift_idx: int,
    expected_distance_mean: float,
    expected_p_value: float,
    expected_likelihood: float,
) -> None:
    """Test streaming mSPRT test on data callback.

    :param X_ref_multivariate: reference multivariate data
    :type X_ref_multivariate: numpy.ndarray
    :param X_test_multivariate: test multivariate data
    :type X_test_multivariate: numpy.ndarray
    :param detector_class: detector distance
    :type detector_class: BaseDataDriftBatch
    :param expected_drift_idx: expected drift index
    :type expected_drift_idx: int
    :param expected_distance_mean: expected distance mean value
    :type expected_distance_mean: float
    :param expected_p_value: expected p-value value
    :type expected_p_value: float
    :param expected_likelihood: expected likelihood value
    :type expected_likelihood: float
    """
    np.random.seed(seed=31)

    alpha = 0.01

    detector = detector_class(  # type: ignore
        callbacks=mSPRT(
            alpha=alpha,
            sigma=1,
            tau=1,
            lambda_=32,
        ),
        window_size=5,
    )
    _ = detector.fit(X=X_ref_multivariate)

    drift_idx = -1
    for i, sample in enumerate(X_test_multivariate, start=1):
        value, callbacks_logs = detector.update(value=sample)
        if value is not None:
            if callbacks_logs["mSPRT"]["p_value"] < alpha:
                drift_idx = i
                break

    assert drift_idx == expected_drift_idx
    assert np.isclose(callbacks_logs["mSPRT"]["distance_mean"], expected_distance_mean)
    assert np.isclose(callbacks_logs["mSPRT"]["p_value"], expected_p_value)
    assert np.isclose(callbacks_logs["mSPRT"]["likelihood"], expected_likelihood)



"""ChiSquareTest (Chi-square test) module."""

import collections
from typing import Optional, List, Tuple, Union

import numpy as np  # type: ignore
from scipy.stats import chi2_contingency  # type: ignore

from frouros.callbacks.base import BaseCallback
from frouros.detectors.data_drift.base import CategoricalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (  # type: ignore
    BaseStatisticalTest,
    StatisticalResult,
)


class ChiSquareTest(BaseStatisticalTest):
    """ChiSquareTest (Chi-square test) [pearson1900x]_ detector.

    :References:

    .. [pearson1900x] Pearson, Karl.
        "X. On the criterion that a given system of deviations from the probable in the
        case of a correlated system of variables is such that it can be reasonably
        supposed to have arisen from random sampling."
        The London, Edinburgh, and Dublin Philosophical Magazine and Journal of
        Science 50.302 (1900): 157-175.
    """

    def __init__(
        self, callbacks: Optional[Union[BaseCallback, List[BaseCallback]]] = None
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[Callback, List[BaseCallback]]]
        """
        super().__init__(
            data_type=CategoricalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    def _statistical_test(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        f_exp, f_obs = self._calculate_frequencies(X_ref=X_ref, X=X)
        statistic, p_value, _, _ = chi2_contingency(
            observed=np.array([f_obs, f_exp]), **kwargs
        )

        test = StatisticalResult(statistic=statistic, p_value=p_value)
        return test

    @staticmethod
    def _calculate_frequencies(
        X_ref: np.ndarray, X: np.ndarray  # noqa: N803
    ) -> Tuple[List[int], List[int]]:
        X_ref_counter, X_counter = [  # noqa: N806
            *map(collections.Counter, [X_ref, X])  # noqa: N806
        ]
        possible_values = set([*X_ref_counter.keys()] + [*X_counter.keys()])
        f_exp, f_obs = {}, {}
        for value in possible_values:
            f_exp[value] = X_ref_counter.get(value, 0)
            f_obs[value] = X_counter.get(value, 0)
        f_exp_values, f_obs_values = [
            *map(list, [f_exp.values(), f_obs.values()])  # type: ignore
        ]
        return f_exp_values, f_obs_values  # type: ignore



"""Data drift detection methods init."""

from .batch import (  # noqa: F401
    BhattacharyyaDistance,
    ChiSquareTest,
    CVMTest,
    EMD,
    HellingerDistance,
    HINormalizedComplement,
    JS,
    KL,
    KSTest,
    PSI,
    MMD,
    WelchTTest,
)

from .streaming import IncrementalKSTest, MMD as MMDStreaming  # noqa: N811

__all__ = [
    "BhattacharyyaDistance",
    "ChiSquareTest",
    "CVMTest",
    "EMD",
    "HellingerDistance",
    "HINormalizedComplement",
    "IncrementalKSTest",
    "JS",
    "KL",
    "KSTest",
    "PSI",
    "MMDStreaming",
    "WelchTTest",
]



"""Welch's T-test module."""

from typing import Optional, List, Union

import numpy as np  # type: ignore
from scipy.stats import ttest_ind  # type: ignore

from frouros.callbacks.base import BaseCallback
from frouros.detectors.data_drift.base import NumericalData, UnivariateData
from frouros.detectors.data_drift.batch.statistical_test.base import (
    BaseStatisticalTest,
    StatisticalResult,
)


class WelchTTest(BaseStatisticalTest):
    """Welch's T-test [welch1947generalization]_ detector.

    :References:

    .. [welch1947generalization] Welch, Bernard L.
        "The generalization of ‘STUDENT'S’problem when several different population
        varlances are involved."
        Biometrika 34.1-2 (1947): 28-35.
    """

    def __init__(
        self, callbacks: Optional[Union[BaseCallback, List[BaseCallback]]] = None
    ) -> None:
        """Init method.

        :param callbacks: callbacks
        :type callbacks: Optional[Union[Callback, List[BaseCallback]]]
        """
        super().__init__(
            data_type=NumericalData(),
            statistical_type=UnivariateData(),
            callbacks=callbacks,
        )

    def _statistical_test(
        self, X_ref: np.ndarray, X: np.ndarray, **kwargs  # noqa: N803
    ) -> StatisticalResult:
        test = ttest_ind(
            a=X_ref, b=X, equal_var=False, alternative="two-sided", **kwargs
        )
        test = StatisticalResult(statistic=test.statistic, p_value=test.pvalue)
        return test



"""Callbacks init."""

from .batch import PermutationTestOnBatchData, ResetOnBatchDataDrift
from .streaming import History, mSPRT, WarningSamplesBuffer

__all__ = [
    "History",
    "mSPRT",
    "PermutationTestOnBatchData",
    "ResetOnBatchDataDrift",
    "WarningSamplesBuffer",
]
